use eyre::Result;
use once_cell::sync::Lazy;
use reqwest::{Response, StatusCode, header::{CONTENT_TYPE, CONTENT_LENGTH, DATE}};
use rsa::RsaPrivateKey;
use serde::{de::DeserializeOwned, Serialize};
use tracing::error;

use crate::{
    signature::{build_signature, format_date},
    FoxError,
};

use super::{SIGNATURE_HEADER, USER_HEADER, SERVER_HEADER};

static APP_USER_AGENT: &str = concat!(env!("CARGO_PKG_NAME"), "/", env!("CARGO_PKG_VERSION"),);

static CLIENT: Lazy<reqwest::Client> = Lazy::new(|| {
    reqwest::Client::builder()
        .user_agent(APP_USER_AGENT)
        .build()
        .unwrap()
});

pub fn is_valid_domain(domain: &str) -> bool {
    addr::parse_domain_name(domain).is_ok()
}

pub async fn get<R: DeserializeOwned>(
    private_key: &RsaPrivateKey,
    self_domain: &str,
    host: &str,
    path: &str,
    user_id: Option<String>,
) -> Result<R> {
    let (signature, date) = build_signature(
        private_key,
        host,
        path,
        None,
        user_id.clone(),
    );

    let mut req = CLIENT
        .get(format!("https://{host}{path}"))
        .header(DATE, format_date(date))
        .header(SIGNATURE_HEADER, signature)
        .header(SERVER_HEADER, self_domain);

    req = if let Some(id) = user_id {
        req.header(USER_HEADER, id)
    } else {
        req
    };

    let resp = req.send().await?;
    handle_response(resp).await
}

pub async fn post<T: Serialize, R: DeserializeOwned>(
    private_key: &RsaPrivateKey,
    self_domain: &str,
    host: &str,
    path: &str,
    user_id: Option<String>,
    body: &T,
) -> Result<R> {
    let body = serde_json::to_string(body)?;

    let (signature, date) = build_signature(
        private_key,
        host,
        path,
        Some(body.len()),
        user_id.clone(),
    );

    let mut req = CLIENT
        .post(format!("https://{host}{path}"))
        .header(DATE, format_date(date))
        .header(SIGNATURE_HEADER, signature)
        .header(SERVER_HEADER, self_domain)
        .header(CONTENT_TYPE, "application/json; charset=utf-8")
        .header(CONTENT_LENGTH, body.len().to_string())
        .body(body);

    req = if let Some(id) = user_id {
        req.header(USER_HEADER, id)
    } else {
        req
    };

    let resp = req.send().await?;
    handle_response(resp).await
}

async fn handle_response<R: DeserializeOwned>(resp: Response) -> Result<R> {
    if resp.status() != StatusCode::OK {
        error!("federation request failed with status code {}", resp.status());
        return Err(FoxError::ResponseNotOk.into());
    }

    let parsed = resp.json::<R>().await?;
    Ok(parsed)
}
